freeslot(
	"sfx_sphfir",
	"sfx_sphhit",
	"sfx_sphkil",
	"sfx_sphjmp",
	"sfx_sphlnd",
	"sfx_sphhov",
	"sfx_sphgru",
	"sfx_sphsel",
	"sfx_sphbad",
	"sfx_sphlav",
	"SPR_SHED",
	"S_SOPHIAHEAD",
	"S_SOPHIADEBRIS",
	"MT_SOPHIAHEAD",
	"MT_SOPHIADEBRIS"
)

sfxinfo[sfx_sphkil].flags = SF_X4AWAYSOUND
sfxinfo[sfx_sphgru].singular = true

local speen = {}
local lasttarget = {}
local loadedthething = false

local beamres = CV_RegisterVar({
	name = "sph_beamresolution",
	defaultvalue = "0.25",
	flags = CV_SHOWMODIF|CV_FLOAT,
	PossibleValue = {MIN = 1, MAX = FRACUNIT-1},
	func = nil
})

local togglefire = CV_RegisterVar({
	name = "sph_togglefire",
	defaultvalue = "Off",
	flags = CV_SHOWMODIF|CV_NETVAR,
	PossibleValue = CV_OnOff
})

addHook("MapChange",function()
	if not isserver
	or loadedthething
		return end
	loadedthething = true
		
	local file = io.openlocal("client/sophia.dat","r")
	
	if not file
		print("No Sophia config found")
		return end
	
	local val = file:read("*n")
	file:close()
	
	if type(val) == "number"
		CV_Set(CV_FindVar("sph_togglefire"),val)
		print("Got Sophia data")
		S_StartSound(nil,sfx_sphgru,consoleplayer)
	else
		print("Error getting Sophia data: expected number, got " .. type(val) .. " (" .. val .. ")")
	end
end)

addHook("GameQuit",function(quitting)
	if not isserver
		return end
		
	local file = io.openlocal("client/sophia.dat","w")
	file:write(togglefire.value)
	print("Successfully saved value " .. tostring(togglefire.value) .. " to Sophia config file")
	file:close()
end)

local headFunc = function(mo,t,p)	
	if not (mo and t and p and mo.valid and t.valid and p.sophia)
		return 
	end
		
	if (t.state >= S_PLAY_SUPER_TRANS1
	and t.state <= S_PLAY_SUPER_TRANS6)
	or p.powers[pw_carry] == CR_NIGHTSMODE
		mo.sprite = SPR_NULL
		return end

	mo.sprite = SPR_SHED
		
	local momz = t.momz
	if mo.eflags & MFE_VERTICALFLIP
		momz = -$
	end
	
	mo.angle = t.angle
	local offset = FRACUNIT*4-min(max(momz,-FRACUNIT*4),FRACUNIT*4)
	if p.sophia.extraheadoff
		offset = min(FRACUNIT*8,max(0,$+p.sophia.extraheadoff))
		if not (p.sophia.extraheadoff / FU)
			p.sophia.extraheadoff = 0
		elseif p.sophia.extraheadoff > 0
			p.sophia.extraheadoff = $-FU
		else
			p.sophia.extraheadoff = $+FU
		end
	end
	
	if mo.eflags & MFE_VERTICALFLIP
		mo.z = $-offset
	else
		mo.z = $+offset
	end
	
	mo.dispoffset = 3
	if p.sophia.target and p.sophia.target.valid
		local pst = p.sophia.target
		mo.angle = R_PointToAngle2(mo.x,mo.y,pst.x,pst.y)
	end
end

local dealSophiaDamage = function(targ,p)
	local off = 12*p.mo.scale
	if p.mo.eflags & MFE_VERTICALFLIP
		off = -$
	end
	
	for i = 1,FRACUNIT,fixdiv(10*FRACUNIT-beamres.value,max(1,R_PointToDist2(p.mo.x,p.mo.y,targ.x,targ.y))) do
		local thok = P_SpawnMobj(ease.linear(i,p.mo.x,targ.x),ease.linear(i,p.mo.y,targ.y),ease.linear(i,p.mo.z+off,targ.z+(targ.height/2)),MT_THOK)
		thok.color = p.skincolor
		thok.scale = p.mo.scale / 2
		thok.destscale = thok.scale / 4
	end
	
	if targ.flags2 & MF2_FRET
	or (targ.player and (targ.player.powers[pw_flashing] or targ.player.powers[pw_invulnerability]))
		return end
		
	if not targ.sophiamaxhealth
		if targ.flags & MF_BOSS
			targ.sophiamaxhealth = 8
		else
			targ.sophiamaxhealth = 3
		end
	end
		
	if targ.state == S_METALSONIC_FLOAT
	or targ.player
		P_InstaThrust(targ,R_PointToDist2(p.mo.x,p.mo.y,targ.x,targ.y),0)
		p.sophia.hitSomeoneNow = true
		--p.sophia.firetime = 2
		--targ.state = S_METALSONIC_STUN
		--targ.flags = ($&~MF_SLIDEME)|MF_GRENADEBOUNCE|MF_ENEMY|MF_SHOOTABLE
	elseif not targ.sophiadamage
		targ.sophiadamage = 1
		S_StartSound(targ,sfx_sphhit)
		p.sophia.hitSomeoneNow = true
	elseif targ.sophiadamage < (targ.sophiamaxhealth - 1)
		targ.sophiadamage = $+1
		S_StartSound(targ,sfx_sphhit)
		p.sophia.hitSomeoneNow = true
	else
		S_StartSound(targ,sfx_sphkil)
		p.sophia.hitSomeoneNow = true
		targ.sophiadamage = 0
		if targ.tracer and targ.tracer.type == MT_EGGSHIELD
			P_KillMobj(targ.tracer,nil,p.mo)
			S_StartSound(targ,sfx_wbreak)
		elseif targ.flags & MF_MONITOR
			P_KillMobj(targ,nil,p.mo)
		else
			P_DamageMobj(targ,nil,p.mo)
			if p.sophia.gear != 3
				S_StartSound(nil,sfx_itemup,p)
				P_GivePlayerRings(p,1)
			end
		end
		if targ and targ.valid
		and targ.flags & MF_BOSS
			P_GivePlayerRings(p,1)
		end
	end
end

local sophiaGearShift = function(p)
	local gear = p.sophia.gear
	
	if gear == 0
		p.accelstart = 64
		p.normalspeed = 32*FRACUNIT
		p.actionspd = 28*FRACUNIT
		p.jumpfactor = FRACUNIT*85/100
		p.sophia.range = 600*FRACUNIT
		p.sophia.firerate = 12
		p.sophia.lavaresist = 10
	elseif gear == 1
		p.accelstart = 64
		p.normalspeed = 36*FRACUNIT
		p.actionspd = 34*FRACUNIT
		p.jumpfactor = FRACUNIT
		p.sophia.range = 800*FRACUNIT
		p.sophia.firerate = 10
		p.sophia.lavaresist = 7
	elseif gear == 2
		p.accelstart = 96
		p.normalspeed = 36*FRACUNIT
		p.actionspd = 40*FRACUNIT
		p.jumpfactor = FRACUNIT
		p.sophia.range = 1000*FRACUNIT
		p.sophia.firerate = 8
		p.sophia.lavaresist = 4
	else
		p.accelstart = 128
		p.normalspeed = 40*FRACUNIT
		p.actionspd = 46*FRACUNIT
		p.jumpfactor = FRACUNIT*1355/1000
		p.sophia.range = 1200*FRACUNIT
		p.sophia.firerate = 5
		p.sophia.lavaresist = 1
	end
end

local sophiaShiftDown = function(p)
	if p.realtime > 1
		S_StartSound(p.mo,sfx_kc65)	
	end
end

local sophiaShiftUp = function(p)
	if p.realtime > 1
		S_StartSound(p.mo,sfx_sphgru)
	end
end

local sophiaCheckGear = function(p)
	// sophiaCheckGear
	// Arguments:
	//	- player_t p
	// Returns: boolean (whether or not another gear shift could have occurred)
	local r = p.rings
	local g = p.sophia.gear
	p.sophia.lastcheckedgear = r
	
	if r == 0
	and g != 0
		p.sophia.gear = 0
		sophiaGearShift(p)
		sophiaShiftDown(p)
	elseif g == 0
		if r >= 12
			p.sophia.gear = 1
			sophiaGearShift(p)
			sophiaShiftUp(p)
			return true
		end
	elseif g == 1
		if r < 8
			p.sophia.gear = 0
			sophiaGearShift(p)
			sophiaShiftDown(p)
		elseif r >= 32
			p.sophia.gear = 2
			sophiaGearShift(p)
			sophiaShiftUp(p)
			return true
		end
	elseif g == 2
		if r < 24
			p.sophia.gear = 1
			sophiaGearShift(p)
			sophiaShiftDown(p)
			return true
		elseif r >= 80
			p.sophia.gear = 3
			sophiaGearShift(p)
			sophiaShiftUp(p)
		end
	else
		p.sophia.gear = 3
		if r < 78
			p.sophia.gear = 2
			sophiaGearShift(p)
			sophiaShiftDown(p)
			return true
		end
	end
	
	return false
end

local initSophia = function() return {
	target = nil,
	lastinput = nil,
	lastcheckedgear = -1,
	gear = -1,
	range = 0,
	firerate = TICRATE/3,
	lavaresist = TICRATE/2,
	hitSomeoneNow = false,
	extraheadoff = 0,
	didjumpbounce = false,
	toggledfire = false
} end

--------------------------------------------------------------------------------

addHook("PostThinkFrame",function()
	for p in players.iterate
		if p.mo and p.mo.valid and p.mo.skin == "sophia" and p.followmobj and p.followmobj.valid
			headFunc(p.followmobj,p.mo,p)
		end
	end
end)

addHook("PlayerSpawn",function(p)
	if skins[p.skin].name == "sophia"
		p.sophia = initSophia()
		if p.sopcarryovergear
			p.sophia.gear = p.sopcarryovergear
			p.sopcarryovergear = nil
			sophiaGearShift(p)
		else
			p.sophia.gear = -1
		end
		p.sophia.lastcheckedgear = -1
	end
end)

addHook("PlayerThink",function(p)
	if p.powers[pw_carry] == CR_NIGHTSMODE
		return end
		
	local mo = p.mo
	
	if not (mo and mo.valid)
		return end
	
	if mo.skin != "sophia"
		p.sophia = nil
		return
	elseif not p.sophia
		p.sophia = initSophia()
	end
	
	p.powers[pw_strong] = $|STR_FLOOR|STR_SPIKE|STR_HEAVY
	if (p.sophia.gear == 3
	and p.speed >= p.normalspeed / 2)
	or p.sophia.metalpunch
		p.powers[pw_strong] = $|STR_PUNCH
	else
		p.powers[pw_strong] = $&~STR_PUNCH
	end
		
	if p.sophia.lastinput == nil
		p.sophia.lastinput = p.cmd.buttons
	end
	
	if (abs(mo.z-mo.ceilingz) < 48*FU and ~mo.eflags & MFE_VERTICALFLIP)
	or (abs(mo.z-mo.floorz) < 48*FU and mo.eflags & MFE_VERTICALFLIP)
		p.sophia.extraheadoff = -12*FU
		
	elseif not P_IsObjectOnGround(mo)
		if (p.powers[pw_carry] == CR_GENERIC
		or p.powers[pw_carry] == CR_PLAYER
		or p.powers[pw_carry] == CR_ROPEHANG
		or p.powers[pw_carry] == CR_PTERABYTE)
		or (p.pflags & PF_JUMPED
		and not p.sophia.didjumpbounce)
		or p.climbing
			p.sophia.extraheadoff = 12*FU
		end
		p.sophia.didjumpbounce = true
	elseif p.sophia.didjumpbounce
		p.sophia.extraheadoff = -6*FU
		S_StartSound(mo,sfx_sphlnd)
		p.sophia.didjumpbounce = false
	end
	
	if p.sophia.gear == -1
		p.sophia.gear = 0
		sophiaGearShift(p)
		p.sophia.lastcheckedgear = p.rings
	elseif p.rings != p.sophia.lastcheckedgear
		if p.rings - p.sophia.lastcheckedgear >= 100
		and p.realtime > 1
			p.rings = $-100
		end
		while sophiaCheckGear(p) do end
	end
	
	if togglefire.value
		if p.cmd.buttons & BT_SPIN
		and ~p.sophia.lastinput & BT_SPIN
			p.sophia.toggledfire = not $
			S_StartSound(nil,sfx_sphsel,p)
		end
	else
		p.sophia.toggledfire = false
	end
		
	local targ = nil
	local targ2 = nil
	local range = p.sophia.range
	
	searchBlockmap("objects",function(r,f)
		local objectdir = R_PointToAngle2(r.x,r.y,f.x,f.y)
		local rvec = {cos(r.angle),sin(r.angle)}
		local fvec = {2*cos(objectdir),2*sin(objectdir)}
		
		if not f.sophianotarget
		and not p.exiting
		and p.playerstate == PST_LIVE
		and P_CheckSight(r,f)
		and (f.flags & (MF_ENEMY|MF_SHOOTABLE|MF_MONITOR)
		or f.state == S_METALSONIC_FLOAT)
		and not (f.flags & (MF_NOCLIP|MF_NOCLIPTHING))
		and f != r
		and f != r.followmobj
		and R_PointToDist2(r.x,r.y,f.x,f.y) <= range
		and (not (targ and targ.valid)
		or R_PointToDist2(r.x,r.y,f.x,f.y) < R_PointToDist2(r.x,r.y,targ.x,targ.y))
		and f.health
		and not (f.player and (gametyperules & GTR_FRIENDLY or (gametyperules & GTR_TEAMS and f.player.team == p.team)))
		--and fixmul(rvec[1], fvec[1]) + fixmul(rvec[2], fvec[2]) >= 0
			targ = f
		end
	end,mo,mo.x-range,mo.x+range,mo.y-range,mo.y+range)
	
	p.sophia.target = targ
	if targ and targ.valid and targ.health
		
		P_SpawnLockOn(p,targ,S_LOCKON1)
		
		if (p.cmd.buttons & BT_SPIN
		or p.sophia.toggledfire)
		and not p.sophia.firetime
			S_StartSound(mo,sfx_sphfir)
			if targ.type == MT_EGGMAN_BOX
			or targ.type == MT_EGGMAN_GOLDBOX
				S_StartSound(nil,sfx_sphbad,p)
			end
			
			p.sophia.firetime = p.sophia.firerate
			dealSophiaDamage(targ,p)
		end
	end
	
	if p.skidtime
	and p.glidetime
		mo.state = S_PLAY_GLIDE_LANDING
		p.skidtime = min($,1)
	end
	
	if mo.state == S_PLAY_GLIDE
	and not S_SoundPlaying(mo,sfx_sphhov)
		S_StartSound(mo,sfx_sphhov)
	elseif mo.state != S_PLAY_GLIDE
	and S_SoundPlaying(mo,sfx_sphhov)
		S_StopSoundByID(mo,sfx_sphhov)
	end
	
	p.sophia.lastinput = p.cmd.buttons
	if p.sophia.firetime
		p.sophia.firetime = $-1
	end
	
	if p.sophia.lavaresisttics
		p.sophia.lavaresisttics = $-1
	end
	
	if p.sophia.metalpunch
		p.sophia.metalpunch = $-1
	end
end)

hud.add(function(v,p,cam)
	if not p.sophia
	or p.powers[pw_carry] == CR_NIGHTSMODE
	or p.exiting
		return end
		
	local gearstring = p.sophia.gear
	if not gearstring
		gearstring = "0"
	elseif gearstring == 3
		gearstring = "Max"
	end
	
	v.drawString(290,0,"Gear " .. gearstring,V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS|V_PERPLAYER,"center")
	
	if gamestate == GS_LEVEL
	and p.sophia
	and p.sophia.target
	and p.sophia.target.valid
	and p.sophia.target.sprite
	and p.mo and p.mo.valid
		if speen[#p]
		and lasttarget[#p]
		and lasttarget[#p] == p.sophia.target
			speen[#p] = ($+1)%64
		else 
			speen[#p] = 1
			lasttarget[#p] = p.sophia.target
		end
		
		local sprpatch, flip = v.getSpritePatch(p.sophia.target.sprite,p.sophia.target.frame&FF_FRAMEMASK,(speen[#p]/8)+1,0)
		local hitFlash = V_REDMAP
		if p.sophia.hitSomeoneNow
			hitFlash = V_INVERTMAP
		end
		
		if sprpatch
			if flip
				flip = V_FLIP
			else
				flip = 0
			end
			
			local skin = TC_DEFAULT
			if p.sophia.hitSomeoneNow
				skin = TC_ALLWHITE
			end
			
			v.drawScaled(290*FRACUNIT,18*FRACUNIT+(p.sophia.target.height/4),FRACUNIT/4,sprpatch,V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS|V_PERPLAYER|flip,v.getColormap(skin,p.sophia.target.color))
			if p.sophia.target.sprite == SPR_TV1P
				v.drawScaled(290*FRACUNIT,(54*FRACUNIT/4)+(p.sophia.target.height/4),FRACUNIT/4,v.getSprite2Patch("sophia",SPR2_LIFE),V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS|V_PERPLAYER|flip,v.getColormap(skin,p.skincolor))
			end
			
			v.drawString(290,28+(p.sophia.target.height/(4*FRACUNIT)),"TARGET",V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS|V_PERPLAYER,"center")
			if p.sophia.target.sophiadamage
				v.drawString(290,18+(p.sophia.target.height/(4*FRACUNIT)),p.sophia.target.sophiamaxhealth - p.sophia.target.sophiadamage .. "/" .. p.sophia.target.sophiamaxhealth,V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS|V_PERPLAYER|hitFlash ,"center")
			end
		end
		
		p.sophia.hitSomeoneNow = false
	else
		speen[#p] = 0
	end
	
	if togglefire.value
		local autofirey = 8
		if p.sophia.target
		and p.sophia.target.valid
			autofirey = $+30+(p.sophia.target.height/(4*FRACUNIT))
		end
		
		local autofire = "OFF"
		if p.sophia.toggledfire
			autofire = "ON"
		end
		
		v.drawString(290,autofirey,"AF: " .. autofire,V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS|V_PERPLAYER,"center")
	end
end)

addHook("ShouldDamage",function(t,i,s,d,dt)
	if i == nil
	and s == nil
	and t and t.valid 
	and t.player and t.player.sophia
	and dt == DMG_FIRE
	and not (t.player.powers[pw_flashing] or t.player.powers[pw_invulnerability] or t.player.powers[pw_shield])
		if t.player.sophia.lavaresisttics
			return false
		elseif t.player.rings <= 0
			return true
		else
			t.player.sophia.lavaresisttics = t.player.sophia.lavaresist
			t.player.rings = $-1
			S_StartSound(nil,sfx_sphlav,t.player)
			return false
		end
	end
end,MT_PLAYER)

addHook("MobjDeath",function(t,i,s,dt)
	if i.player and i.player.sophia
		i.momx = $*3/4
		i.momy = $*3/4
		i.momz = $*3/4
	end
end,MT_SPIKE)

addHook("MapChange",function()
	for p in players.iterate
		if p.sophia
		and not p.sopcarryoverrings
			p.sopcarryoverrings = p.rings
			p.sopcarryovergear = p.sophia.gear
		end
	end
end)

addHook("MapLoad",function()
	for p in players.iterate
		if p.powers[pw_carry] == CR_NIGHTSMODE
			return end
			
		if p.sophia
			if p.sopcarryoverrings
			and mapheaderinfo[gamemap].startrings
			and p.sopcarryoverrings > mapheaderinfo[gamemap].startrings
				p.rings = p.sopcarryoverrings
				p.sophia.lastcheckedgear = -1 --p.sopcarryoverrings
				--p.sophia.gear = p.sopcarryovergear
				--sophiaCheckGear(p)
			end 
			p.sopcarryoverrings = nil
		end
	end
end)

addHook("MobjDeath",function(t,s,i,dt)
	if t.player
	and t.player.sophia
		P_SpawnMobj(t.x,t.y,t.z,MT_EXPLODE)
		S_StartSound(t,sfx_pop)
		
		local debris = P_SpawnMobj(t.x,t.y,t.z-FU,MT_SOPHIADEBRIS)
		debris.dispoffset = 3
		debris.eflags = t.eflags
		debris.angle = R_PointToAngle(t.x,t.y) + ANGLE_180
		debris.momz = t.momz
		P_InstaThrust(debris,debris.angle+ANGLE_90,P_RandomRange(-10,10)*FU)
		
		debris = P_SpawnMobj(t.x,t.y,t.z-FU,MT_SOPHIADEBRIS)
		debris.color = t.color
		debris.frame = C
		debris.dispoffset = 3
		debris.eflags = t.eflags
		debris.angle = R_PointToAngle(t.x,t.y) + ANGLE_180
		debris.momz = t.momz
		P_InstaThrust(debris,debris.angle+ANGLE_90,-3*FU)
		
		debris = P_SpawnMobj(t.x,t.y,t.z-FU,MT_SOPHIADEBRIS)
		debris.color = t.color
		debris.frame = D
		debris.dispoffset = 3
		debris.eflags = t.eflags
		debris.angle = R_PointToAngle(t.x,t.y) + ANGLE_180
		debris.momz = t.momz
		P_InstaThrust(debris,debris.angle+ANGLE_90,3*FU)
		
		if not netgame or splitscreen
			t.player.sopcarryoverrings = -1
		end
	end
end,MT_PLAYER)

addHook("MobjThinker",function(mo)
	mo.sophianotarget = (mo.state == S_EGGGUARD_STND)
end,MT_EGGGUARD)

local inverthealth = function(mo) 
	if mo.flags & MF_BOSS
		mo.sophiamaxhealth = 3 
	else
		mo.sophiamaxhealth = 8
	end
end

addHook("MobjSpawn",inverthealth,MT_EGGMAN_BOX)
addHook("MobjSpawn",inverthealth,MT_EGGMAN_GOLDBOX)
addHook("MobjSpawn",inverthealth,MT_1UP_BOX)
addHook("MobjSpawn",inverthealth,MT_RING_BOX)
addHook("MobjSpawn",inverthealth,MT_INVULN_BOX)
addHook("MobjSpawn",inverthealth,MT_SNEAKERS_BOX)
addHook("MobjSpawn",inverthealth,MT_METALSONIC_BATTLE)

local mwahaha = function(t,i,s,dt)
	if s and s.valid and s.player and s.player.sophia
		S_StartSound(nil,sfx_bewar1,s.player)
	end
end

addHook("MobjDeath",mwahaha,MT_EGGMAN_BOX)
addHook("MobjDeath",mwahaha,MT_EGGMAN_GOLDBOX)

addHook("MobjThinker",function(mo)
	for p in players.iterate
		if p.sophia
		and p.mo and p.mo.valid
		and R_PointToDist2(mo.x,mo.y,p.mo.x,p.mo.y) < 128*FU
		and not P_IsObjectOnGround(p.mo)
			p.sophia.metalpunch = 2
		end
	end
end,MT_METALSONIC_BATTLE)